/*
 Copyright (C) 2013 Stefano Zanotti @ C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CELL_H_
#define CELL_H_

#include "fmatdef.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <string.h>
#include "bitWrite.h"
#include "Text.h"
#include "LedDriver.h"
#include "Painter.h"
#include "Fonts.h"


class __FlashStringHelper;

namespace FlashMat_lite {

class FlashMat {

public:
    FlashMat(); // do not call this! use the predefined instance 'cell'

    void setup();

	void fill(const uint8_t color[3]);

    // if 0, uses the default text (stored inside the cell itself)
	void drawText(const Text* text = 0);
    
    const Text& getText() const;
	void setText(const char string[]);
	void setText(const uint8_t string[], uint8_t len);
	void setText(const __FlashStringHelper* str);
	void setTextFont(uint8_t fontNum);
	void setTextColor(const uint8_t color[3]);
	void setTextBgColor(const uint8_t color[3]);
	void setTextOverlay(bool overlay);
    void setTextPosition(int16_t x, int16_t y);
    void centerText();
	
	void drawPixel(uint8_t x, uint8_t y, const uint8_t color[3]);
	void drawLineH(uint8_t x0, uint8_t x1, uint8_t y, const uint8_t color[3]);
	void drawLineV(uint8_t x, uint8_t y0, uint8_t y1, const uint8_t color[3]);
	void drawLine(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, const uint8_t color[3]);
	void drawRect(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, const uint8_t color[3], bool filled);
    
    void drawGradient(const uint8_t color1[3], const uint8_t color2[3], Gradient type);

	void copyBuffer(); // copy front to back, to perform the next drawings over it
    
    void swapBuffers(); // swaps the front and back buffers, i.e. shows the image that has just been drawn on
    void show(); // friendlier name for swapBuffers()
    
    uint8_t getPixel(uint8_t x, uint8_t y, uint8_t colorPlane) const;
    uint8_t getPixelShown(uint8_t x, uint8_t y, uint8_t colorPlane) const;
    
private:
    // front and back buffers, 3 colors, N rows, M columns, B-bits color depth
    uint8_t _frmBuffs[2][FRAME_SIZE]; // used as two matrices [3][ROWS][COLS] (manually-calculated indexing) (this is true for B=8, otherwise the bits are packed)

    volatile uint8_t _frontBuff; // select which one of _frmBuffs is the buff that is currently shown
    volatile uint8_t _backBuff; // select which one of _frmBuffs is the buff that is currently being drawn on

    // all text-related settings
    Text _text;
};

// only instance of the class
extern FlashMat fm;


// IMPLEMENTATION
// Placed here to get the opportunity to inline it
// (it's made mostly of delegations)
/////////////////////////////////////////////////////////////////////

inline
FlashMat::FlashMat() :
        _frontBuff(0),
        _backBuff(1),
        _text()
{
    // reset both buffers to black
    memset(_frmBuffs, 0, sizeof(_frmBuffs));

    driver.frameBuffer = _frmBuffs[_frontBuff];
}

inline
void FlashMat::setup() {
    cli(); // disable interrupts
    driver.setup();
    sei(); // enable interrupts
}

inline
void FlashMat::swapBuffers() {
    _backBuff = 1 - _backBuff;
    _frontBuff = 1 - _backBuff;

    driver.frameBuffer = _frmBuffs[_frontBuff];
}

inline
void FlashMat::show() {
    swapBuffers();
}

inline
void FlashMat::fill(const uint8_t color[3]) {
    Painter::fill(_frmBuffs[_backBuff], color);
}

inline
void FlashMat::copyBuffer() {
    Painter::copyBuffer(_frmBuffs[_backBuff], _frmBuffs[_frontBuff]);
}

inline
const Text& FlashMat::getText() const {
    return _text;
}

inline
void FlashMat::setText(const char string[]) {
    _text.str[0] = '\0';
    strncat(_text.str, string, TEXT_BUFFER_SIZE); // safer than strcpy, appends a null character and copies at most N characters
}

inline
void FlashMat::setText(const uint8_t string[], uint8_t len) {
  #if TEXT_BUFFER_SIZE < 255
    if(len > TEXT_BUFFER_SIZE) {
        len = TEXT_BUFFER_SIZE;
    }
  #endif
    
    memcpy(_text.str, string, len);
    _text.str[len] = '\0';
}

inline
void FlashMat::setText(const __FlashStringHelper* str) {
    for(uint16_t i = 0; i < TEXT_BUFFER_SIZE; i++) {
        _text.str[i] = pgm_read_byte(reinterpret_cast<const uint8_t*>(str) + i);
        
        if(_text.str[i] == '\0')
        return;
    }
}

inline
void FlashMat::setTextFont(uint8_t fontNum) {
    _text.fontId = fontNum;
    _text.font = Fonts::getFont(fontNum);
}

inline
void FlashMat::setTextColor(const uint8_t color[3]) {
    memcpy(_text.color, color, 3);
}

inline
void FlashMat::setTextOverlay(bool overlay) {
    _text.overlay = overlay;
}

inline
void FlashMat::setTextBgColor(const uint8_t color[3]) {
    memcpy(_text.bgColor, color, 3);
}

inline
void FlashMat::drawText(const Text* text) {
    Painter::drawText(_frmBuffs[_backBuff], (text ? *text : _text));
}

inline
void FlashMat::setTextPosition(int16_t x, int16_t y) {
    _text.topleft[X] = x;
    _text.topleft[Y] = y;
}

inline
void FlashMat::centerText() {
    _text.center();
}

inline
void FlashMat::drawPixel(uint8_t x, uint8_t y, const uint8_t color[3]) {
    Painter::drawPixel(_frmBuffs[_backBuff], x, y, color);
}

inline
void FlashMat::drawLineH(uint8_t x0, uint8_t x1, uint8_t y, const uint8_t color[3]) {
    Painter::drawLineH(_frmBuffs[_backBuff], x0, x1, y, color);
}

inline
void FlashMat::drawLineV(uint8_t x, uint8_t y0, uint8_t y1, const uint8_t color[3]) {
    Painter::drawLineV(_frmBuffs[_backBuff], x, y0, y1, color);
}

inline
void FlashMat::drawLine(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, const uint8_t color[3]) {
    Painter::drawLine(_frmBuffs[_backBuff], x0, y0, x1, y1, color);
}

inline
void FlashMat::drawRect(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, const uint8_t color[3], bool filled) {
    Painter::drawRect(_frmBuffs[_backBuff], x0, y0, x1, y1, color, filled);
}

inline
void FlashMat::drawGradient(const uint8_t color1[3], const uint8_t color2[3], Gradient type) {
    Painter::drawGradient(_frmBuffs[_backBuff], color1, color2, type);
}

inline
uint8_t FlashMat::getPixel(uint8_t x, uint8_t y, uint8_t colorPlane) const {
    if(x >= MATRIX_COLS || y >= MATRIX_ROWS || colorPlane >= 3) {
        return 0;
    }
    
    return Painter::getPixel(_frmBuffs[_backBuff], colorPlane, x, y);
}

inline
uint8_t FlashMat::getPixelShown(uint8_t x, uint8_t y, uint8_t colorPlane) const {
    if(x >= MATRIX_COLS || y >= MATRIX_ROWS || colorPlane >= 3) {
        return 0;
    }
    
    return Painter::getPixel(_frmBuffs[_frontBuff], colorPlane, x, y);
}

} // namespace

#endif /* CELL_H_ */
