/*
 Copyright (C) 2013 Stefano Zanotti @ C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "LedDriver.h"

#include <binary.h>

// interrupt on timer1
//ISR(TIMER1_OVF_vect){ // overflow
ISR(TIMER1_COMPA_vect) { // CTC
    FlashMat_lite::driver.showAndAdvanceRow(); //TODO
}

namespace FlashMat_lite {

LedDriver::LedDriver() :
        frameBuffer(0),
        _currRow(0)
{
}

void LedDriver::setup() {
    initPins();
    initTimer();
}

void LedDriver::initPins() const {
    BSET1(DI_DDR, DI_PIN); // output
    BSET1(DCKI_DDR, DCKI_PIN); // output
    BSET1(EN_DDR, EN_PIN); // output
    BSET1(B0_DDR, B0_PIN); // output
    BSET1(B1_DDR, B1_PIN); // output
    BSET1(B2_DDR, B2_PIN); // output

#if USE_GCKI
    // Setting for grayscale clock as output.
    // This is needed to actually have the clock signal on the pin
    // (if it's set as an input pin, no waveform is generated on the pin,
    // even if the timer is enabled)
    BSET1(GCKI_DDR, GCKI_PIN);
#endif

    // set all outputs to 0
    BSET0(DI_PORT, DI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);
    BSET0(B0_PORT, B0_PIN);
    BSET0(B1_PORT, B1_PIN);
    BSET0(B2_PORT, B2_PIN);

#if USE_GCKI
    // This value is output to the pin whenever the timer is disabled (via disableGCK()).
    // By setting it now as low, we are sure it is in the low state when disabled
    // (actually, the MY9221 accepts a high steady state as well)
    BSET0(GCKI_PORT, GCKI_PIN);
#endif

    disableDecoder();
}

void LedDriver::initTimer() const {
    // We refresh the rows at 800Hz (1250us); having 8 rows, the whole display
    // is refreshed at 100Hz.
    // (Refresh rates below 80Hz create quite psychedelic effects, especially on
    // peripheral vision, so we avoid them).
    //
    // NOTE: the frequency is fixed by the interrupt, but the time consumed
    // at each iteration by the driver depends on many configuration parameters
    // and on compilation options, too, so they may vary by great amounts.

#if MATRIX_ROWS != 8
    #error "The refresh timer only supports 8 rows"
#endif

    // Timer for RGB line advance
    /////////////////////////////
    TCCR1A = 0; // reset timer settings
    TCCR1B = 0; // reset timer settings
    TIMSK1 |= _BV(OCIE1A); // enable CTC interrupt on Timer1
    TCCR1B |= _BV(CS11); // prescaler /8
    TCCR1B |= _BV(WGM12); // enable CTC (Clear Timer on Compare match)
    TCNT1 = 0; // reset timer counter
    OCR1A = 2499; // set CTC value (100Hz)
    
#if USE_GCKI
    // Timer for Grayscale clock (settings for 8MHz)
    /////////////////////////////
    TCCR2A = 0;// reset timer settings
    TCCR2B = 0;// reset timer settings
    TIMSK2 = 0;// no interrupts
    TCCR2A |= _BV(COM2B0);// toggle OC2B on compare match
    TCCR2A |= _BV(WGM21);// enable CTC (Clear Timer on Compare match)
    TCNT2 = 0;// reset timer counter
    OCR2B = 0;// set CTC value (0 = reset on every clock cycle; this setting plus toggle on compare -> frequency = f_clk / 2)
    // must also enable the timer (using a prescaler /1, setting CS20), but we do that only when needed, with enableGCK()
#endif
}

#if USE_GCKI
    inline void LedDriver::enableGCK() const {
        TCCR2B |= _BV(CS20); // prescaler /1
    }
#endif

#if USE_GCKI
    inline void LedDriver::disableGCK() const {
        TCCR2B &= ~_BV(CS20); // disable timer (assuming that it was prescaled at /1; see enableGCK())
    }
#endif

inline void LedDriver::enableDecoder() const {
    BSET1(EN_PORT, EN_PIN);
}

inline void LedDriver::disableDecoder() const {
    BSET0(EN_PORT, EN_PIN);
}

inline void LedDriver::sendRowAddress() const {
    BSET(_currRow & 0x01, B0_PORT, B0_PIN);
    BSET(_currRow & 0x02, B1_PORT, B1_PIN);
    BSET(_currRow & 0x04, B2_PORT, B2_PIN);
}


inline void LedDriver::delay_latchStart() const {
    // Delay from last data to first latch pulse [Tstart]
    // 220us <= T
    //
    // From experiments, it seems that 8-9us are sufficient
    delayUs(10);
}

inline void LedDriver::delay_latchHigh() const {
    // Min duration of latch high state [twH(DI)]
    // 70ns <= T
    
#if F_CPU == 16000000L
    // + 2 clock cycles to set the pin state (125ns)
    //
    // Nothing more needed
#else
    #error "Unhandled F_CPU"
#endif
}

inline void LedDriver::delay_latchLow() const {
    // Min duration of latch low state [twL(DI)]
    // 230ns <= T
    
#if F_CPU == 16000000L
    // + 2 clock cycles to set the pin state (125ns)
    __asm__ __volatile__ (
        "nop" "\n\t"
        "nop"
    );
#else
    #error "Unhandled F_CPU"
#endif
}

inline void LedDriver::delay_latchEnd() const {
    // Delay from last latch pulse to next data [Tstop]
    // 200ns + N*10ns <= T   (N is the number of cascaded devices; e.g. 0 if only device)
    // better: 200ns + N*12ns <= T   (see tpLH4, which is tipically 12ns)
    
#if F_CPU == 16000000L
    //  +2 clock cycles to set the pin state
    //  0~ 4 drivers (1~2 matrices): 250  ns  ->  4-2 = 2 cycles
    //  5~10 drivers (3~5 matrices): 312.5ns  ->  5-2 = 3 cycles
    // 11~15 drivers (6~8 matrices): 375  ns  ->  6-2 = 4 cycles
    
#if MATRIX_COLS <= 16
    nop(); nop();
#elif MATRIX_COLS <= 40
    nop(); nop(); nop();
#elif MATRIX_COLS <= 64
    nop(); nop(); nop(); nop();
#else
    #error "Unhandled number of columns"
#endif
    
#else
    #error "Unhandled F_CPU"
#endif
}

inline void LedDriver::delay_dcki() const {
    // Min duration of clock pulse (high or low) [tw(DCK)]
    // 50us <= T <= 7.2us
    //
    // Using direct port manipulation and F_CPU >= 8MHz, the upper bound
    // is certainly not a problem.
    
#if F_CPU == 16000000L
    // Less than 1 clock cycle: nothing needed
#else
    #error "Unhandled F_CPU"
#endif
}

inline void LedDriver::delay_di_su() const {
    // Min duration of data set up time [tsu(D)]
    // 10us <= T
    
#if F_CPU == 16000000L
    // Less than 1 clock cycle: nothing needed
#else
    #error "Unhandled F_CPU"
#endif
}

inline void LedDriver::delay_di_h() const {
    // Min duration of data hold time [th(D)]
    // 10us <= T
    
#if F_CPU == 16000000L
    // Less than 1 clock cycle: nothing needed
#else
    #error "Unhandled F_CPU"
#endif
}

inline
void LedDriver::advanceRowIndex() {
    _currRow = (_currRow + (MATRIX_ROWS/4 + 1)) % MATRIX_ROWS;
}

inline
void LedDriver::turnOffLeds() const {

#if 0  // READABLE VERSION
    for (int j = 0; j < MATRIX_COLS / 4; ++j) { // each MY9221 drives 4 pixels
        sendLedCmd();
        BSET0(DI_PORT, DI_PIN);
        // send 192 0's
        for (uint8_t i = 0; i < 192; ++i) {
            BTGL(DCKI_PORT, DCKI_PIN);
        }
    }

    // send latch, but without enabling the row
    // If we re-enabled the row right now, the previous color would still
    // be shown (until the current PWM-frame has been completed)
    sendLatch(true);
#else // if ! readable version
    for (uint16_t j = MATRIX_COLS / 4 + 1 ; --j > 0;) { // each MY9221 drives 4 pixels
        sendLedCmd();
        BSET0(DI_PORT, DI_PIN);
        // send 192 0's
        for (uint8_t i = 24 + 1; --i > 0;) { // 24 iters * 8 toggles = 192 clock pulses
            BSET1(DCKI_PORT, DCKI_PIN);
            BSET0(DCKI_PORT, DCKI_PIN);
            BSET1(DCKI_PORT, DCKI_PIN);
            BSET0(DCKI_PORT, DCKI_PIN);
            BSET1(DCKI_PORT, DCKI_PIN);
            BSET0(DCKI_PORT, DCKI_PIN);
            BSET1(DCKI_PORT, DCKI_PIN);
            BSET0(DCKI_PORT, DCKI_PIN);
        }
    }

    // send latch, but without enabling the row
    // If we re-enabled the row right now, the previous color would still
    // be shown (until the current PWM-frame has been completed)
    sendLatch(true);
#endif // if readable version
}

inline
void LedDriver::sendLedCmd() const {
    /*                  COMMAND BITS 15:0  (for MY9221 LED driver)
     * ==============================================================================
     * 15:11 |  Temp     | Not used Please filled with all "0"
     * ==============================================================================
     *   10  |  hspd     | Iout Tr/Tf select
     *       |           |     0 : Iout slow mode
     *       |           |     1 : Iout fast mode
     * ==============================================================================
     *  9:8  |  bs[1:0]  | Grayscale resolution select
     *       |           |     00 : 8-bit grayscale application
     *       |           |     01 : 12-bit grayscale application
     *       |           |     10 : 14-bit grayscale application
     *       |           |     11 : 16-bit grayscale application
     * ==============================================================================
     *  7:5  |  gck[2:0] | Internal oscillator freq. select
     *       |           |     000 : original freq (8.6MHz)
     *       |           |     001 : original freq/2
     *       |           |     010 : original freq/4
     *       |           |     011 : original freq/8
     *       |           |     100 : original freq/16
     *       |           |     101 : original freq/64
     *       |           |     110 : original freq/128
     *       |           |     111 : original freq/256
     *       |           |  If CMD[3]=1, please set CMD[7:5]=000
     * ==============================================================================
     *   4   |   sep     |  Output waveform select
     *       |           |     0 : MY-PWM output waveform
     *       |           |          (similar to traditional waveform)
     *       |           |     1 : APDM output waveform
     * ==============================================================================
     *   3   |   osc     |  Grayscale clock source select
     *       |           |     0 : internal oscillator (8.6MHz) (internal GCK source)
     *       |           |     1 : external clock from GCKI pin (external GCK source)
     * ==============================================================================
     *   2   |   pol     |  Output polarity select
     *       |           |     0 : work as LED driver
     *       |           |     1 : work as MY-PWM/APDM generator
     * ==============================================================================
     *   1   |   cntset  |  Counter reset select
     *       |           |     0 : free running mode
     *       |           |            (for every pwm frame, the last received data
     *       |           |            is shown; when new data arrives, the current
     *       |           |            frame is executed completely, before switching
     *       |           |            to the new data)
     *       |           |     1 : counter reset mode (Only usable when osc = "1")
     *       |           |            The behavior is defined by CMD[0].
     * ==============================================================================
     *   0   |   onest   |  One-shot select (Only usable when cntset = "1")
     *       |           |     0 : frame cycle repeat mode
     *       |           |            (for every pwm frame, the last received data
     *       |           |            is shown; when new data arrives, the current
     *       |           |            frame is immediately stopped, and a new frame
     *       |           |            is started with the new data)
     *       |           |     1 : frame cycle One-shot mode
     *       |           |            (when new data arrives, one single pwm frame
     *       |           |            is executed with this data; then the display
     *       |           |            is turned off)
     * ==============================================================================
     */

    #if USE_GCKI
        // LED driver command bits:
        //  00000  Temp (reserved)
        //      1  hspd              fast mode
        //     00  bs[1-0]           8-bit grayscale (with higher bit depths we experienced flickering problems: the rows were cyclically too bright, causing a "sliding" artifact on the image)
        //    000  gck[2-0]          original frequency (8.6 MHz)
        //      1  sep               APDM waveform
        //      1  osc               use external grayscale clock
        //      0  pol               LED driver (instead of PWM-generator)
        //      1  cntset            frame cycle repeat mode
        //      0  onest             frame cycle repeat mode
        const uint16_t cmd = 0b0000010000011010;
    #else // USE_GCKI == false
        // LED driver command bits:
        //  00000  Temp (reserved)
        //      1  hspd              fast mode
        //     00  bs[1-0]           8-bit grayscale (with higher bit depths we experienced flickering problems: the rows were cyclically too bright, causing a "sliding" artifact on the image)
        //    000  gck[2-0]          original frequency (8.6 MHz)
        //      1  sep               APDM waveform
        //      0  osc               use internal oscillator
        //      0  pol               LED driver (instead of PWM-generator)
        //      0  cntset            free running mode
        //      0  onest             free running mode
        const uint16_t cmd = 0b0000010000010000;
    #endif // if USE_GCKI

    #if 0 // READABLE VERSION
        uint16_t cmd = _ledDriverCmd;
        for (uint8_t i = 16; i > 0; --i) {
            BSET(cmd & 0x8000, DI_PORT, DI_PIN); // send bit
            BTGL(DCKI_PORT, DCKI_PIN);// clock pulse (data is read on both clock edges)
            cmd <<= 1;// prepare next bit
        }
    #else
        BSET(cmd & 0x8000, DI_PORT, DI_PIN);// send bit
        BSET1(DCKI_PORT, DCKI_PIN);// clock pulse (data is read on both clock edges)
        BSET(cmd & 0x4000, DI_PORT, DI_PIN);
        BSET0(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x2000, DI_PORT, DI_PIN);
        BSET1(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x1000, DI_PORT, DI_PIN);
        BSET0(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0800, DI_PORT, DI_PIN);
        BSET1(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0400, DI_PORT, DI_PIN);
        BSET0(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0200, DI_PORT, DI_PIN);
        BSET1(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0100, DI_PORT, DI_PIN);
        BSET0(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0080, DI_PORT, DI_PIN);
        BSET1(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0040, DI_PORT, DI_PIN);
        BSET0(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0020, DI_PORT, DI_PIN);
        BSET1(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0010, DI_PORT, DI_PIN);
        BSET0(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0008, DI_PORT, DI_PIN);
        BSET1(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0004, DI_PORT, DI_PIN);
        BSET0(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0002, DI_PORT, DI_PIN);
        BSET1(DCKI_PORT, DCKI_PIN);
        BSET(cmd & 0x0001, DI_PORT, DI_PIN);
        BSET0(DCKI_PORT, DCKI_PIN);
    #endif
}

inline
void LedDriver::sendLedData(uint8_t val) const {
    // Always send 16 bits of data: 8 times '0' and then the 8 bits of the color
    
    // Timing constraints (as of MY9221 datasheet)
    // + Both high and low states of the DCKI must be <7.2us
    //   With 16MHz, this means <= 115 clock cycles

#if 0 // READABLE VERSION

    // send 8 low bits
    BSET0(DI_PORT, DI_PIN);
    for (int i = 0; i < 8; i++) {
        BTGL(DCKI_PORT, DCKI_PIN);
    }

    // send grayscale data
    for (uint8_t i = 0; i < 8; i++) {
        BSET(val & 0x80, DI_PORT, DI_PIN); // send bit
        BTGL(DCKI_PORT, DCKI_PIN);// clock pulse
        val <<= 1;// prepare next bit
    }
#else // if // readable version

    // send 8 low bits
    BSET0(DI_PORT, DI_PIN);
    BSET1(DCKI_PORT, DCKI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);
    BSET1(DCKI_PORT, DCKI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);
    BSET1(DCKI_PORT, DCKI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);
    BSET1(DCKI_PORT, DCKI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);

    // SEND GRAYSCALE DATA
    // for lower color depth, the lowest bits are duplicate of the highest ones,
    // to achieve full "swing" 0-255
    BSET(val & 0x80, DI_PORT, DI_PIN); // send bit
    BSET1(DCKI_PORT, DCKI_PIN); // clock pulse
    BSET(val & 0x40, DI_PORT, DI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);
    BSET(val & 0x20, DI_PORT, DI_PIN);
    BSET1(DCKI_PORT, DCKI_PIN);
    BSET(val & 0x10, DI_PORT, DI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);
    BSET(val & 0x08, DI_PORT, DI_PIN);
    BSET1(DCKI_PORT, DCKI_PIN);
    BSET(val & 0x04, DI_PORT, DI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);
    BSET(val & 0x02, DI_PORT, DI_PIN);
    BSET1(DCKI_PORT, DCKI_PIN);
    BSET(val & 0x01, DI_PORT, DI_PIN);
    BSET0(DCKI_PORT, DCKI_PIN);
#endif
}

inline
void LedDriver::sendLatch(bool wait) const {
    // SEND LATCH PULSES TO THE LED DRIVERS

    // Timing constraints (as of MY9221 datasheet)
    // + T_start: from last bit on DI (last DCKI pulse) from first latch pulse (on DI),
    //   there must be > 220us. It seems this constraint can be relaxed a lot (see below)
    // + Latch pulse must be >70ns (high state) and >230ns (low state)
    //   With 16MHz this means >=2 clock cycles and >=4 clock cycles
    // + T_stop: from last latch pulse to next DI data (DCKI pulse), there must be
    //   > 220ns + N*10ns (N=number of cascaded MY9221).
    //   With 16MHz this means >=4 clock cycles (for 1-4 total devices) and >= 5 (for 5-10 devices)
    //
    // Constraints for external grayscale clock:
    // + Must have been low for >100ns when the second falling edge of the latch pulses arrives
    // + Must stay low for >300ns after the third falling edge of the latch pulses has arrived
    // These constraints are surely met (as long as we met the constraints on the latch pulses)
    // if we keep the grayscale clock low from the first falling edge to the fourth falling edge

    // due to the timing constraints, little or no optimization can be made on the latch pulses

    BSET0(DI_PORT, DI_PIN);    // to return to 0 if the last data bit was 1

    if(wait) {
        delay_latchStart();
    }

    // disable decoder so that the display doesn't flicker during the change
    // (note: if the drivers are set to finish their current window frame
    // before switching to the new data, part of the current row's color
    // may be shown in the next row, as we switch to it; this effect
    // is clearly visible; it is highly reduced by using the GCKI, though).
    disableDecoder();

    // send 4 clock pulses, and temporarily stop the grayscale clock
    BSET1(DI_PORT, DI_PIN); delay_latchHigh();
    BSET0(DI_PORT, DI_PIN); delay_latchLow();

#if USE_GCKI
    disableGCK();
#endif

    BSET1(DI_PORT, DI_PIN); delay_latchHigh();
    BSET0(DI_PORT, DI_PIN); delay_latchLow();
    BSET1(DI_PORT, DI_PIN); delay_latchHigh();
    BSET0(DI_PORT, DI_PIN); delay_latchLow();
    BSET1(DI_PORT, DI_PIN); delay_latchHigh();
    BSET0(DI_PORT, DI_PIN); delay_latchLow();

#if USE_GCKI
    enableGCK();
#endif

	// This delay is probably unnecessary (before sending new data we must at least
    // prepare the new data); we do it just to be extra sure about the timings
    delay_latchEnd();

    //NOTE: there is still a very faint color leakage from one row to the next one, but this
    // is only visible at a very close distance (about 20cm), only if it happens on very dark
    // pixels (almost black) and only if the viewer is actually looking for such a phenomenon.
    // Moreover: some leakage seems to happen on some columns, for every row (not only for the next
    // one); this happens only on some columns and on some colours, and something similar
    // happens with other firmware as well (e.g. Rainbowduino v3).
}

inline
void LedDriver::sendLatchAndRowAddress(bool wait) const {
    sendLatch(wait); // this also disables the decoder, must reenable it when the proper row has been selected

    sendRowAddress();
    enableDecoder();
}

void LedDriver::showAndAdvanceRow() {
if (frameBuffer == 0) {
        // invalid pointer: just turn off the display
        turnOffLeds();
        return;

    } else {

    #if USE_GCKI
        // To drive the MY9221 we need these fundamental steps:
        // + Send the led data (together with commands to configure the MY9221 behavior)
        // + Wait a non-negligible amount of time, needed by the MY9221 itself
        // + Send a latch command to make the MY9221 switch to the new data
        //
        // To avoid wasting the time spent in waiting, we can avoid the trivial
        // behavior:
        //   + Enter ISR
        //   + Send data, wait, send latch
        //   + Exit ISR
        // and use this one instead:
        //   + Enter ISR
        //   + Send latch for the previous ISR
        //   + Send data for this ISR
        //   + Exit ISR
        // This way the wait disappears (the CPU is free to do calculations during that time).
        // We have a delay between the data and the latch (about 1 ISR cycle), but this guarantees
        // the delay is long enough, and it is not noticeable (the delay is f 1 cycle, which
        // must be much faster than the frame rate: it cannot be perceived).
        // Finally, to avoid flashing random lights during the very first ISR (when the MY9221
        // has not been sent data, i.e. they contain random data), we do not send the latch
        // during the first ISR call.
        
        // To avoid the flash of one line the first time this ISR is run (the MY9221 do not contain
        // meaningful data yet)
        static bool first = true;
        
        if(!first) {
            // send latch pulses to store the data into the drivers,
            // and enable the proper row via the decoder
            sendLatchAndRowAddress(false);
        }
        
        // Advance the row index, for the next call to this function;
        // advance in a non-sequential way to avoid "sliding in" effects
        advanceRowIndex();
        
        // start from the rightmost pixel (the first one to be shifted out)
        uint8_t* const baseptr = frameBuffer + MATRIX_ROW_SIZE * _currRow + MATRIX_ROW_SIZE;
        uint8_t* pixR = baseptr + R * FRAME_SIZE_1_COLOR;
        uint8_t* pixG = baseptr + G * FRAME_SIZE_1_COLOR;
        uint8_t* pixB = baseptr + B * FRAME_SIZE_1_COLOR;
        // note: the previous pointers point after the last byte; this is useful
        // for sendFrame(), which can decrement this pointer at every iteration,
        // without handling the first iteration in a special manner

	    sendFrame(pixR, pixG, pixB);
        
	    first = false;
    
    #else
        // No GCKI: we must turn the matrix off before turning it on with the new row color.
        // This is because we cannot synchronize perfectly with the MY9221's internal "frame cycle"
        // (one complete cycle of PWM/ADPM current drive); this causes the old cycle to continue
        // while we have already switched to the new row, causing color leakage (clearly
        // visible when color leaks to a row that ought to be black).
        // To avoid this, we first color the row in black, and then color it with the new color:
        // this way we still have color leakage, but it's always black that leaks (i.e. we have at
        // most a faint dim in brightness, much less annoying than a true color leakage).
        // This problem does not appear when we use GCKI: with it, we can use another mode of the MY9221,
        // ("counter reset mode" instead of "free running mode") that causes the "frame cycle" to restart
        // as soon as new data has been latched in.
        uint8_t* const baseptr = frameBuffer + MATRIX_ROW_SIZE * _currRow + MATRIX_ROW_SIZE;
        uint8_t* pixR = baseptr + R * FRAME_SIZE_1_COLOR;
        uint8_t* pixG = baseptr + G * FRAME_SIZE_1_COLOR;
        uint8_t* pixB = baseptr + B * FRAME_SIZE_1_COLOR;
    
        turnOffLeds(); // to avoid color leakage from one row to the next one
        sendFrame(pixR, pixG, pixB);
        sendLatchAndRowAddress(true);
        advanceRowIndex();
    #endif
    }
}

/*
 * RAINBOWDUINO PIN MAPPING
 *
 * In software, columns and rows ranges in 0~7, increasing from top-left to bottom-right.
 *
 * Sequential RGB: one driver drives 8 reds and 4 greens, the other 8 blues and 4 greens
 * (MY9221's output specs are disregarded).
 *
 * MY9221 uses these letters: A=red, B=green, C=blue.
 *
 * We use common anode 2088 RGB matrices:
 *
 *                       COLUMNS (RGB, RGB, ...)
 *             1     2     3     4     5     6     7     8
 *          [ 9]  [10]  [11]  [12]  [13]  [14]  [15]  [16]
 *           [28]  [27]  [26]  [25]  [24]  [23]  [22]  [21]
 *            [ 1]  [ 2]  [ 3]  [ 4]  [ 5]  [ 6]  [ 7]  [ 8]
 *   1 [17]
 *   2 [18]
 * R 3 [19]
 * O 4 [20]
 * W 5 [29]
 * S 6 [30]
 *   7 [31]
 *   8 [32]
 *
 * The matrix pinout is (viewed from TOP):
 *    _________________
 *    |  32 31 ... 17 |
 *    |               |
 *    |               |
 *    |o  1  2 ... 16 |
 *    -----------------
 *
 * The pixel-to-pin mapping is as follows: (note: colors and anodes ("COM") span in 1~8 here)
 * (Note: with BRGB matrices, R and B are swapped)
 * =========================================
 * | PIXEL |   MY9221   | MATRIX | DECODER |
 * =========================================
 * | COM 2 |     -      |   17   | Y0 [15] |
 * | COM 3 |     -      |   18   | Y1 [14] |
 * | COM 3 |     -      |   19   | Y2 [13] |
 * | COM 4 |     -      |   20   | Y3 [12] |
 * | COM 5 |     -      |   29   | Y4 [11] |
 * | COM 6 |     -      |   30   | Y5 [10] |
 * | COM 7 |     -      |   31   | Y6 [ 9] |
 * | COM 8 |     -      |   32   | Y7 [ 7] |
 * =========================================
 * |   R3  | OUTA0 [ 6] |    9   | First MY9221 (driven by MCU)
 * |   R2  | OUTB0 [ 5] |   28   |
 * |   R1  | OUTC0 [ 4] |    1   |
 * |   R6  | OUTA1 [ 9] |   10   |
 * |   R5  | OUTB1 [ 8] |   27   |
 * |   R4  | OUTC1 [ 7] |    2   |
 * |   G1  | OUTA2 [18] |   11   |
 * |   R8  | OUTB2 [17] |   26   |
 * |   R7  | OUTC2 [16] |    3   |
 * |   G4  | OUTA3 [21] |   12   |
 * |   G3  | OUTB3 [20] |   25   |
 * |   G2  | OUTC3 [19] |    4   |
 * ===============================
 * |   G7  | OUTA4 [ 6] |   13   | Second MY9221 (cascaded, driven by the first MY9221)
 * |   G6  | OUTB4 [ 5] |   24   |
 * |   G5  | OUTC4 [ 4] |    5   |
 * |   B2  | OUTA5 [ 9] |   14   |
 * |   B1  | OUTB5 [ 8] |   23   |
 * |   G8  | OUTC5 [ 7] |    6   |
 * |   B5  | OUTA6 [18] |   15   |
 * |   B4  | OUTB6 [17] |   22   |
 * |   B3  | OUTC6 [16] |    7   |
 * |   B8  | OUTA7 [21] |   16   |
 * |   B7  | OUTB7 [20] |   21   |
 * |   B6  | OUTC7 [19] |    8   |
 * ===============================
 */

/*
 * FLASHMAT PIN MAPPING (for single 8x8 module; others are cascaded replications of it)
 *
 * In software, columns and rows ranges in 0~7, increasing from top-left to bottom-right.
 *
 * Interlaced RGB: each driver drives 4 RGB leds.
 *
 * MY9221 uses these letters: A=red, B=green, C=blue.
 *
 * We use common anode 2088 RGB matrices:
 *
 *                       COLUMNS (RGB, RGB, ...)
 *             1     2     3     4     5     6     7     8
 *          [ 9]  [10]  [11]  [12]  [13]  [14]  [15]  [16]
 *           [28]  [27]  [26]  [25]  [24]  [23]  [22]  [21]
 *            [ 1]  [ 2]  [ 3]  [ 4]  [ 5]  [ 6]  [ 7]  [ 8]
 *   1 [17]
 *   2 [18]
 * R 3 [19]
 * O 4 [20]
 * W 5 [29]
 * S 6 [30]
 *   7 [31]
 *   8 [32]
 *
 * The matrix pinout is (viewed from TOP):
 *    _________________
 *    |  32 31 ... 17 |
 *    |               |
 *    |               |
 *    |o  1  2 ... 16 |
 *    -----------------
 *
 * The pixel-to-pin mapping is as follows:
 * (Note: with BRGB matrices, R and B are swapped)
 * =========================================
 * | PIXEL |   MY9221   | MATRIX | DECODER |
 * =========================================
 * |ANODE 0|     -      |   17   | Y0 [15] |
 * |ANODE 1|     -      |   18   | Y1 [14] |
 * |ANODE 2|     -      |   19   | Y2 [13] |
 * |ANODE 3|     -      |   20   | Y3 [12] |
 * |ANODE 4|     -      |   29   | Y4 [11] |
 * |ANODE 5|     -      |   30   | Y5 [10] |
 * |ANODE 6|     -      |   31   | Y6 [ 9] |
 * |ANODE 7|     -      |   32   | Y7 [ 7] |
 * =========================================
 * |   R0  | OUTA0 [ 6] |    9   | First MY9221 (driven by MCU)
 * |   G0  | OUTB0 [ 5] |   28   |
 * |   B0  | OUTC0 [ 4] |    1   |
 * |   R1  | OUTA1 [ 9] |   10   |
 * |   G1  | OUTB1 [ 8] |   27   |
 * |   B1  | OUTC1 [ 7] |    2   |
 * |   R2  | OUTA2 [18] |   11   |
 * |   G2  | OUTB2 [17] |   26   |
 * |   B2  | OUTC2 [16] |    3   |
 * |   R3  | OUTA3 [21] |   12   |
 * |   G3  | OUTB3 [20] |   25   |
 * |   B3  | OUTC3 [19] |    4   |
 * ===============================
 * |   R4  | OUTA4 [ 6] |   13   | Second MY9221 (cascaded, driven by the first MY9221)
 * |   G4  | OUTB4 [ 5] |   24   |
 * |   B4  | OUTC4 [ 4] |    5   |
 * |   R5  | OUTA5 [ 9] |   14   |
 * |   G5  | OUTB5 [ 8] |   23   |
 * |   B5  | OUTC5 [ 7] |    6   |
 * |   R6  | OUTA6 [18] |   15   |
 * |   G6  | OUTB6 [17] |   22   |
 * |   B6  | OUTC6 [16] |    7   |
 * |   R7  | OUTA7 [21] |   16   |
 * |   G7  | OUTB7 [20] |   21   |
 * |   B7  | OUTC7 [19] |    8   |
 * ===============================
 */

inline
void LedDriver::sendFrame(const uint8_t* pixR, const uint8_t* pixG, const uint8_t* pixB) const {
    
    #if LED_MATRIX_TYPE == LED_MATRIX_BSRG
        // everything's fine
    #elif LED_MATRIX_TYPE == LED_MATRIX_BRGB
        // must swap red and blue
        const uint8_t* tmp = pixR;
        pixR = pixB;
        pixB = tmp;
    #endif

#if 0 // READABLE VERSION (for 8bit depth, which needs no bits unpacking)
    #if PIN_MAPPING == PINMAP_RAINBOWDUINO
        // command (second MY9221)
        // 8 blue  pixels (rightmost-to-leftmost)
        // 4 green pixels (rightmost-to-middle)
        // command (first MY9221)
        // 4 green pixels (middle-to-leftmost)
        // 8 red   pixels (rightmost-to-leftmost)

        sendLedCmd();
    
        for (int i = 0; i < 8; ++i) {
            sendLedData(*(--pixB));
        }
        for (int i = 0; i < 4; ++i) {
            sendLedData(*(--pixG));
        }

        sendLedCmd();
    
        for (int i = 0; i < 4; ++i) {
            sendLedData(*(--pixG));
        }
        for (int i = 0; i < 8; ++i) {
            sendLedData(*(--pixR));
        }

    #elif PIN_MAPPING == PINMAP_FLASHMAT
        // for each 8x8 module (on a single line; i.e. only 8 rows allowed)
        // (starting from the rightmost)
        for (int m = 0; m < MATRIX_COLS / 8; ++m) {
            // command (second MY9221)
            // 4 R,G,B pixels (rightmost-to-middle)
            // command (first MY9221)
            // 4 R,G,B pixels (middle-to-leftmost)

            for (int j = 0; j < 2; ++j) {
                sendLedCmd();
                for (int i = 0; i < 4; ++i) {
                    sendLedData(*(--pixR));
                    sendLedData(*(--pixG));
                    sendLedData(*(--pixB));
                }
            }
        }
    #endif
#endif

    #if PIN_MAPPING == PINMAP_RAINBOWDUINO && COLOR_DEPTH == 8
        sendLedCmd();
        
        sendLedData(*--pixB); // blue pixels from 7 to 0
        sendLedData(*--pixB);
        sendLedData(*--pixB);
        sendLedData(*--pixB);
        sendLedData(*--pixB);
        sendLedData(*--pixB);
        sendLedData(*--pixB);
        sendLedData(*--pixB);
        
        sendLedData(*--pixG); // green pixels from 7 to 0
        sendLedData(*--pixG);
        sendLedData(*--pixG);
        sendLedData(*--pixG);

        sendLedCmd();
        
        sendLedData(*--pixG);
        sendLedData(*--pixG);
        sendLedData(*--pixG);
        sendLedData(*--pixG);
        
        sendLedData(*--pixR); // red pixels from 7 to 0
        sendLedData(*--pixR);
        sendLedData(*--pixR);
        sendLedData(*--pixR);
        sendLedData(*--pixR);
        sendLedData(*--pixR);
        sendLedData(*--pixR);
        sendLedData(*--pixR);
        
    #elif PIN_MAPPING == PINMAP_RAINBOWDUINO && COLOR_DEPTH == 4
        uint8_t v;

        sendLedCmd();

        v = *--pixB;    sendLedData(unpack<1>(v)); // blue pixels from 7 to 0
                        sendLedData(unpack<0>(v));
        v = *--pixB;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
        v = *--pixB;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
        v = *--pixB;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));

        v = *--pixG;    sendLedData(unpack<1>(v)); // green pixels from 7 to 0
                        sendLedData(unpack<0>(v));
        v = *--pixG;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));

        sendLedCmd();

        v = *--pixG;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
        v = *--pixG;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));

        v = *--pixR;    sendLedData(unpack<1>(v)); // red pixels from 7 to 0
                        sendLedData(unpack<0>(v));
        v = *--pixR;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
        v = *--pixR;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
        v = *--pixR;    sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
    
    #elif PIN_MAPPING == PINMAP_RAINBOWDUINO && COLOR_DEPTH == 2
        uint8_t v;

        sendLedCmd();

        v = *--pixB;    sendLedData(unpack<3>(v)); // blue pixels from 7 to 0
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
        v = *--pixB;    sendLedData(unpack<3>(v));
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));

        v = *--pixG;    sendLedData(unpack<3>(v)); // green pixels from 7 to 0
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));

        sendLedCmd();

        v = *--pixG;    sendLedData(unpack<3>(v));
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));

        v = *--pixR;    sendLedData(unpack<3>(v)); // red pixels from 7 to 0
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
        v = *--pixR;    sendLedData(unpack<3>(v));
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
    
    #elif PIN_MAPPING == PINMAP_RAINBOWDUINO && COLOR_DEPTH == 1
        uint8_t v;

        sendLedCmd();

        v = *--pixB;    sendLedData(unpack<7>(v)); // blue pixels from 7 to 0
                        sendLedData(unpack<6>(v));
                        sendLedData(unpack<5>(v));
                        sendLedData(unpack<4>(v));
                        sendLedData(unpack<3>(v));
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));

        v = *--pixG;    sendLedData(unpack<7>(v)); // green pixels from 7 to 0
                        sendLedData(unpack<6>(v));
                        sendLedData(unpack<5>(v));
                        sendLedData(unpack<4>(v));

        sendLedCmd();

                        sendLedData(unpack<3>(v));
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));

        v = *--pixR;    sendLedData(unpack<7>(v)); // red pixels from 7 to 0
                        sendLedData(unpack<6>(v));
                        sendLedData(unpack<5>(v));
                        sendLedData(unpack<4>(v));
                        sendLedData(unpack<3>(v));
                        sendLedData(unpack<2>(v));
                        sendLedData(unpack<1>(v));
                        sendLedData(unpack<0>(v));
    
    #elif PIN_MAPPING == PINMAP_FLASHMAT && COLOR_DEPTH == 8
        for (uint8_t i = 0; i < MATRIX_COLS / 4; ++i) {
            sendLedCmd();

            sendLedData(*--pixR);
            sendLedData(*--pixG);
            sendLedData(*--pixB);

            sendLedData(*--pixR);
            sendLedData(*--pixG);
            sendLedData(*--pixB);

            sendLedData(*--pixR);
            sendLedData(*--pixG);
            sendLedData(*--pixB);

            sendLedData(*--pixR);
            sendLedData(*--pixG);
            sendLedData(*--pixB);
        }
    
    #elif PIN_MAPPING == PINMAP_FLASHMAT && COLOR_DEPTH == 4
        for (uint8_t i = 0; i < MATRIX_COLS / 4; ++i) {
            sendLedCmd();

            uint8_t vr = *--pixR;
            uint8_t vg = *--pixG;
            uint8_t vb = *--pixB;

            sendLedData(unpack<1>(vr));
            sendLedData(unpack<1>(vg));
            sendLedData(unpack<1>(vb));

            sendLedData(unpack<0>(vr));
            sendLedData(unpack<0>(vg));
            sendLedData(unpack<0>(vb));

            vr = *--pixR;
            vg = *--pixG;
            vb = *--pixB;

            sendLedData(unpack<1>(vr));
            sendLedData(unpack<1>(vg));
            sendLedData(unpack<1>(vb));

            sendLedData(unpack<0>(vr));
            sendLedData(unpack<0>(vg));
            sendLedData(unpack<0>(vb));
        }
    
    #elif PIN_MAPPING == PINMAP_FLASHMAT && COLOR_DEPTH == 2
        for (uint8_t i = 0; i < MATRIX_COLS / 4; ++i) {
            sendLedCmd();

            uint8_t vr = *--pixR;
            uint8_t vg = *--pixG;
            uint8_t vb = *--pixB;

            sendLedData(unpack<3>(vr));
            sendLedData(unpack<3>(vg));
            sendLedData(unpack<3>(vb));

            sendLedData(unpack<2>(vr));
            sendLedData(unpack<2>(vg));
            sendLedData(unpack<2>(vb));

            sendLedData(unpack<1>(vr));
            sendLedData(unpack<1>(vg));
            sendLedData(unpack<1>(vb));

            sendLedData(unpack<0>(vr));
            sendLedData(unpack<0>(vg));
            sendLedData(unpack<0>(vb));
        }
    
    #elif PIN_MAPPING == PINMAP_FLASHMAT && COLOR_DEPTH == 1
        for (uint8_t i = 0; i < MATRIX_COLS / 8; ++i) {
            uint8_t vr = *--pixR;
            uint8_t vg = *--pixG;
            uint8_t vb = *--pixB;

            sendLedCmd();

            sendLedData(unpack<7>(vr));
            sendLedData(unpack<7>(vg));
            sendLedData(unpack<7>(vb));

            sendLedData(unpack<6>(vr));
            sendLedData(unpack<6>(vg));
            sendLedData(unpack<6>(vb));

            sendLedData(unpack<5>(vr));
            sendLedData(unpack<5>(vg));
            sendLedData(unpack<5>(vb));

            sendLedData(unpack<4>(vr));
            sendLedData(unpack<4>(vg));
            sendLedData(unpack<4>(vb));

            sendLedCmd();

            sendLedData(unpack<3>(vr));
            sendLedData(unpack<3>(vg));
            sendLedData(unpack<3>(vb));

            sendLedData(unpack<2>(vr));
            sendLedData(unpack<2>(vg));
            sendLedData(unpack<2>(vb));

            sendLedData(unpack<1>(vr));
            sendLedData(unpack<1>(vg));
            sendLedData(unpack<1>(vb));

            sendLedData(unpack<0>(vr));
            sendLedData(unpack<0>(vg));
            sendLedData(unpack<0>(vb));
        }
    
    #endif
}

LedDriver driver;

} /*namespace*/
