/*
 Copyright (C) 2013 Stefano Zanotti @ C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LEDDRIVER_H_
#define LEDDRIVER_H_

/*
 * Firmware to drive an RGB 8x8 LED matrix using two cascaded MY9221
 * (12-channel LED drivers).
 *
 * The two drivers are used to drive one row of the matrix at a time
 * (the matrix is common anode; it must be driven one row at a time).
 *
 * Usage:
 * + call setup()
 * + manually enable interrupts (e.g. using sei()); this firmware uses TIMER1
 *   to cycle through rows
 * + provide the pointer to the frame buffer (where the NxM image is stored);
 *   it must be organized as a uint8_t[3][ROWS][COLS]
 */

/*
 * Implementation details.
 *
 * The LEDs are driven one row at a time using two MY9221.
 * The row is selected with a decoder (3bit-to-8lines), that has an enable pin.
 *
 * The rows are cycled non-sequentially to avoid "sliding in" artifacts, though
 * the refresh rate should be high enough to avoid by itself such artifacts.
 *
 * The rows are refreshed @800Hz, so that the whole image is refreshed @100Hz.
 *
 * The MY9221 are used with 8bit-deep colors (8+8+8), since higher depths caused
 * flickering problems (the rows were cyclically too bright, causing sliding-in
 * artifacts as in old CRT TVs).
 *
 * Since the MY9221 are used in "Free Running mode", when switching from one row
 * to the next one the color may "leak": until the PWM-cycle is complete, the old
 * color is still shown. To avoid this, the MY9221 are first cleared (are sent a
 * completely black frame) and then are sent the new frame, so that it is the black
 * color to leak (not so visible, since at worst it looks like a small decrease in
 * brightness).
 * This clearing is not needed if the grayscale clock is enabled (and connected on
 * the hardware board, which is not the case of the Rainbowduino): the grayscale
 * clock allows us to stop the PWM cycle of the MY9221 during the line switch,
 * greatly reducing the color leakage (which is very faintly perceptible, but
 * only at a few centimeters of distance and only on non-opaque matrices).
 * Since we needn't clear the matrix, the grayscale clock introduces a great
 * speed gain (almost 2x).
 *
 * These pins are used to drive the matrix:
 * + DI and DCKI: the DI and DCKI inputs of the MY9221 (cascaded)
 * + B0, B1, B2, EN: address bits and enable bit of the decoder that drives
 *    the matrix's anodes.
 * + D3: grayscale clock
 *
 * The driver is compatible with both Rainbowduino and FlashMat (you must
 * specify the board in "fmatdefs.h", since their pin mappings are different;
 * see the .cpp file for the pin mapping).
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include "fmatdef.h"
#include "bitWrite.h"
#include "time_utils.h"

namespace FlashMat_lite {

#if PIN_MAPPING == PINMAP_RAINBOWDUINO && (MATRIX_COLS > 8 || MATRIX_ROWS > 8)
    #error "The Rainbowduino hardware only supports 8x8 modules (single matrix)"
#endif

#if MATRIX_ROWS > 8
    #error "The driver does not support >8 rows yet (we need more pins for row selection)"
#endif

// PORT: a set of pins (an 8bit register that holds the value of the pins)
// BIT:  the bit to select one single pin of a port
// DDR:  Data Direction Register (choose between input/output for the each pin of a port)
// Pin-mapping is Rainbowduino-compatible (except SWAP, not used on Rainbowduino)
#if USED_BOARD == BOARD_RAINBOWDUINO_V3 || USED_BOARD == BOARD_FLASHMAT
    #define DI_PORT    PORTB      // LED driver Serial Data Input (out)
    #define DI_PIN     _BV(PORTB0)
    #define DI_DDR     DDRB
    #define DCKI_PORT  PORTB      // LED driver Serial Data Clock (out)
    #define DCKI_PIN   _BV(PORTB1)
    #define DCKI_DDR   DDRB
    #define B0_PORT    PORTD      // Row-selecting decoder bit 0: PIN PD4 (out)
    #define B0_PIN     _BV(PORTD4)
    #define B0_DDR     DDRD
    #define B1_PORT    PORTD      // Row-selecting decoder bit 1: PIN PD5 (out)
    #define B1_PIN     _BV(PORTD5)
    #define B1_DDR     DDRD
    #define B2_PORT    PORTD      // Row-selecting decoder bit 2: PIN PD6 (out)
    #define B2_PIN     _BV(PORTD6)
    #define B2_DDR     DDRD
    #define EN_PORT    PORTD      // Row-selecting decoder enable: PIN PD7 (out)
    #define EN_PIN     _BV(PORTD7)
    #define EN_DDR     DDRD
#else
    #error "Unsupported board"
#endif

#if USE_GCKI
    #define GCKI_PORT   PORTD  // Grayscale clock for led drivers: PIN PD3 (out) (on timer2)
    #define GCKI_PIN    _BV(PORTD3)
    #define GCKI_DDR    DDRD
#endif

class LedDriver {
public:
    LedDriver(); // do not call this! use the predefined instance 'driver'

    void setup();

    // The frame is shown row-by-row, so we must cycle through all the rows
    // at a frequency higher than the human eye can perceive. This function shows the
    // current row and then advance the row index.
    // Must be called on timer ticks.
    void showAndAdvanceRow();

    // front frame buffer
    // pointer to FlashMat's frame buffer (uint8_t[FRAME_SIZE] = uint8_t[3][ROWS][COLS])
    // the buffer content shouldn't change after its pointer has been set here -> no need to be volatile
    uint8_t* frameBuffer;

private:
    // currently lit row (used by timer only -> non-volatile)
    uint8_t _currRow;

    void initPins() const;
    void initTimer() const;

    // send command bits to the LED driver
    void sendLedCmd() const;

    // send color bits to the LED driver
    void sendLedData(uint8_t val) const;

    // send enable signal to the decoder (that selects the current row)
    void enableDecoder() const;
    // send enable signal to the decoder (that selects the current row)
    void disableDecoder() const;
    // send row address to the decoder
    void sendRowAddress() const;

    // Turns off the current row, to avoid its color "leaking" into
    // the next row (since the Free Running mode of MY9221 always
    // completes its current pwm-frame, and can't know that we changed row)
    void turnOffLeds() const;

    // Send latch data and also disables the decoder (must be re-enabled
    // by the caller when the row address has been set)
    // 'wait' is true if we must wait the time needed by the MY9221 before
    // sending the latch pulses, false if we are alredy sure that time has already spassed
    void sendLatch(bool wait) const;

    // Send latch pulses to the LED driver and selects the correct row
    // with the decoder.
    // 'wait' is true if we must wait the time needed by the MY9221 before
    // sending the latch pulses, false if we are alredy sure that time has already spassed
    void sendLatchAndRowAddress(bool wait) const;

#if USE_GCKI
    // Enable the grayscale clock for the drivers
    void enableGCK() const;
#endif

#if USE_GCKI
    // Stops the grayscale clock (used while latching data)
    void disableGCK() const;
#endif


    // To extract the proper bits from the byte (in case the color depth is not 8)
    template<uint8_t num>
    uint8_t unpack(uint8_t color) const __attribute__((always_inline));

    // Send command + current row color bits (i.e. multiple calls to sendLedData)
    // the pointer must point BEYOND the last pixel (this makes the function easier,
    // since it needs to send data backwards: it can then decrement and send, decrement
    // and send, and so on)
    void sendFrame(const uint8_t* pixR, const uint8_t* pixG, const uint8_t* pixB) const;
    
    // delays needed by the MY9221 (different implementations can handle different CPU
    // frequency / architecture) [in square brackets, the name given by MY9221's datasheet]
    void delay_latchStart() const __attribute__((always_inline)); // delay from last data to first latch pulse [Tstart]
    void delay_latchHigh()  const __attribute__((always_inline)); // min duration of latch high state [twH(DI)]
    void delay_latchLow()   const __attribute__((always_inline)); // min duration of latch low state [twL(DI)]
    void delay_latchEnd()   const __attribute__((always_inline)); // delay from last latch pulse to next data [Tstop]
    void delay_dcki()       const __attribute__((always_inline)); // min duration of clock pulse (high or low) [tw(DCK)]  // unused for now (it matters only for freqs >=  20MHz)
    void delay_di_su()      const __attribute__((always_inline)); // min duration of data set up time [tsu(D)]            // unused for now (it matters only for freqs >= 100MHz)
    void delay_di_h()       const __attribute__((always_inline)); // min duration of data hold time [th(D)]               // unused for now (it matters only for freqs >= 100MHz)
        
    void advanceRowIndex();
};

template<uint8_t num>
uint8_t LedDriver::unpack(uint8_t color) const {
// #if COLOR_DEPTH == 8
//     assert(num < 1);
// #elif COLOR_DEPTH == 4
//     assert(num < 2);
// #elif COLOR_DEPTH == 2
//     assert(num < 4);
// #elif COLOR_DEPTH == 1
//     assert(num < 8);
// #endif

    // unpack and replicate on LSBs
    #if COLOR_DEPTH == 8
        return color;
    #elif COLOR_DEPTH == 4
        if (num == 0) {
            return (color & 0xF0) | (color >> 4);
        } else {
            return (color << 4) | (color & 0x0F);
        }
    #elif COLOR_DEPTH == 2
        if (num == 0) {
            return ((color & 0xC0)     )
                 | ((color & 0xC0) >> 2)
                 | ((color & 0xC0) >> 4)
                 | ((color & 0xC0) >> 6);
        } else if (num == 1) {
            return ((color & 0x30) << 2)
                 | ((color & 0x30)     )
                 | ((color & 0x30) >> 2)
                 | ((color & 0x30) >> 4);
        } else if (num == 2) {
            return ((color & 0x0C) << 4)
                 | ((color & 0x0C) << 2)
                 | ((color & 0x0C)     )
                 | ((color & 0x0C) >> 2);
        } else {
            return ((color & 0x03) << 6)
                 | ((color & 0x03) << 4)
                 | ((color & 0x03) >> 2)
                 | ((color & 0x03)     );
        }
    #elif COLOR_DEPTH == 1
        return (color & (0x80 >> num)) ? 0xFF : 0x00;
    #endif
}

extern LedDriver driver;

} // namespace

#endif /* LEDDRIVER_H_ */
