/*
 Copyright (C) 2013 Stefano Zanotti @ C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Painter.h"

namespace FlashMat_lite {

uint8_t Painter::getPixel(const uint8_t buff[FRAME_SIZE], uint8_t colorPlane, uint8_t x, uint8_t y) {
    //assert(x >= 0 && x < MATRIX_COLS && y >= 0 && y <= MATRIX_ROWS);

    const uint16_t index = x + MATRIX_COLS * (y + MATRIX_ROWS * colorPlane);

#if COLOR_DEPTH == 8
    return buff[index];
#elif COLOR_DEPTH == 4
    const uint8_t val = buff[index / 2];
    switch(index % 2) {
        default:
        case 0:
            return val & 0xF0;
        case 1:
            return val << 4;
    }
#elif COLOR_DEPTH == 2
    const uint8_t val = buff[index / 4];
    switch (index % 4) {
        default:
        case 0:
            return val & 0xC0;
        case 1:
            return (val << 2) & 0xC0;
        case 2:
            return (val << 4) & 0xC0;
        case 3:
            return val << 6;
    }
#elif COLOR_DEPTH == 1
    const uint8_t val = buff[index / 8];
    return (val << (index % 8)) & 0x80;
#else
    #error "Unmanaged color depth"
#endif
}

void Painter::setPixel(uint8_t value, uint8_t buff[FRAME_SIZE], uint8_t colorPlane, uint8_t x, uint8_t y) {
    //assert(x >= 0 && x < MATRIX_ROWS && y >= 0 && y <= MATRIX_ROWS);
    
    const uint16_t index = x + MATRIX_COLS * (y + MATRIX_ROWS * colorPlane);

#if COLOR_DEPTH == 8
    buff[index] = value;
#elif COLOR_DEPTH == 4
    uint8_t* p = buff + (index / 2);
    switch(index % 2) {
        case 0:
            *p = (*p & 0x0F) | (value & 0xF0);
            return;
        case 1:
            *p = (*p & 0xF0) | (value >> 4);
            return;
    }
#elif COLOR_DEPTH == 2
    uint8_t* p = buff + (index / 4);
    switch(index % 4) {
        case 0:
            *p = (*p & 0x3F) | (value & 0xC0);
            return;
        case 1:
            *p = (*p & 0xCF) | ((value & 0xC0) >> 2);
            return;
        case 2:
            *p = (*p & 0xF3) | ((value & 0xC0) >> 4);
            return;
        case 3:
            *p = (*p & 0xFC) | (value & 0xC0) >> 6;
            return;
    }
#elif COLOR_DEPTH == 1
    uint8_t* p = buff + (index / 8);
    BSET(value & 0x80, *p, _BV(7 - index % 8));
#else
    #error "Unmanaged color depth"
#endif
}

void Painter::fill(uint8_t buff[FRAME_SIZE], const uint8_t color[3]) {
    uint8_t r = color[R];
    uint8_t g = color[G];
    uint8_t b = color[B];

    // create appropriate pattern (duplicate color) for pixel packing,
    // get pointer to each color plane and then fill it
#if COLOR_DEPTH == 8
    // already ok
#elif COLOR_DEPTH == 4
    r = (r & 0xF0) | (r >> 4);
    g = (g & 0xF0) | (g >> 4);
    b = (b & 0xF0) | (b >> 4);
#elif COLOR_DEPTH == 2
    r = packDuplColor2(r);
    g = packDuplColor2(g);
    b = packDuplColor2(b);
#elif COLOR_DEPTH == 1
    r = (r & 0x80) ? 0xFF : 0x00;
    g = (g & 0x80) ? 0xFF : 0x00;
    b = (b & 0x80) ? 0xFF : 0x00;
#else
    #error "Unmanaged color depth"
#endif

    memset(buff + R * FRAME_SIZE_1_COLOR, r, FRAME_SIZE_1_COLOR);
    memset(buff + G * FRAME_SIZE_1_COLOR, g, FRAME_SIZE_1_COLOR);
    memset(buff + B * FRAME_SIZE_1_COLOR, b, FRAME_SIZE_1_COLOR);
}

void Painter::copyBuffer(uint8_t dst[FRAME_SIZE], const uint8_t src[FRAME_SIZE]){
    memcpy(dst, src, FRAME_SIZE);
}

void Painter::drawPixel(uint8_t buff[FRAME_SIZE],
                        uint8_t x, uint8_t y, const uint8_t color[3]) {
    // This check allows drawLines() and similar to draw primitives outside the buffer,
    // without danger.
    if(/*x >= 0 &&*/ x < MATRIX_COLS && /*y >= 0 &&*/ y < MATRIX_ROWS){
        setPixel(color[R], buff, R, x, y);
        setPixel(color[G], buff, G, x, y);
        setPixel(color[B], buff, B, x, y);
    }
}

void Painter::drawLineH(uint8_t buff[FRAME_SIZE],
                        uint8_t x0, uint8_t x1, uint8_t y, const uint8_t color[3]){
    if(x0 > x1) {
        swap(x0, x1);
    }
    
    if(/*y < 0 ||*/ y >= MATRIX_ROWS) {
        return;
    }
    
    if(x0 >= MATRIX_COLS /*|| x1 < 0*/) {
        return;
    }
    
    /*if (x0 < 0) {
        x0 = 0;
    }*/
    
    if(x1 >= MATRIX_COLS) {
        x1 = MATRIX_COLS - 1;
    }
    
    for(uint8_t x = x0; x <= x1; x++) {
        setPixel(color[R], buff, R, x, y);
        setPixel(color[G], buff, G, x, y);
        setPixel(color[B], buff, B, x, y);
    }
}

void Painter::drawLineV(uint8_t buff[FRAME_SIZE],
                        uint8_t x, uint8_t y0, uint8_t y1, const uint8_t color[3]) {
    if(y0 > y1) {
        swap(y0, y1);
    }
    
    if(/*x < 0 ||*/ x >= MATRIX_COLS) {
        return;
    }
    
    if(y0 >= MATRIX_ROWS /*|| y1 < 0*/) {
        return;
    }
    
    /*if(y0 < 0) {
        y0 = 0;
    }*/
    
    if(y1 >= MATRIX_ROWS) {
        y1 = MATRIX_ROWS-1;
    }
    
    for(uint8_t y = y0; y <= y1; y++) {
        setPixel(color[R], buff, R, x, y);
        setPixel(color[G], buff, G, x, y);
        setPixel(color[B], buff, B, x, y);
    }
}

void Painter::drawLine(uint8_t buff[FRAME_SIZE],
                       uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, const uint8_t color[3]) {
    if(y0 == y1) {
        drawLineH(buff, x0, x1, y0, color);
    } else if(x0 == x1) {
        drawLineV(buff, x0, y0, y1, color);
    } else {
        // Bresenham's algorithm, taken from http: //en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
        bool steep = abs(y1 - y0) > abs(x1 - x0);
        if(steep) {
            swap(x0, y0);
            swap(x1, y1);
        }
        if(x0 > x1) {
            swap(x0, x1);
            swap(y0, y1);
        }
        uint8_t deltax = x1 - x0;
        uint8_t deltay = abs(y1 - y0);
        uint8_t error = deltax / 2;
        int8_t ystep = (y0 < y1) ? 1 : -1;
        uint8_t y = y0;
        for (uint8_t x = x0; x <= x1; x++) {
            if (steep) {
                drawPixel(buff, y, x, color);
            } else {
                drawPixel(buff, x, y, color);
            }
            
            if(error >= deltay){
                error -= deltay;
            } else {
                error += (deltax - deltay);
                y += ystep;
            }
        }
    }
}

void Painter::drawRect(uint8_t buff[FRAME_SIZE],
                       uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, const uint8_t color[3], bool filled) {
    
    if(filled) {
        if(y0 > y1) {
            swap(y0, y1);
        }
        for(uint8_t y = y0; y <= y1; y++) {
            drawLineH(buff, x0, x1, y, color);
        }
    } else {
        drawLineH(buff, x0, x1, y0, color);
        drawLineH(buff, x0, x1, y1, color);
        drawLineV(buff, x0, y0, y1, color);
        drawLineV(buff, x1, y0, y1, color);
    }
}

void Painter::drawGradient(uint8_t buff[FRAME_SIZE],
                           const uint8_t color1[3], const uint8_t color2[3], Gradient type) {
    switch (type) {
        case GRAD_WEST:
        case GRAD_NORTH:
            drawGradientStraight(buff, color1, color2, type == GRAD_NORTH);
            break;
        case GRAD_NORTH_WEST:
        case GRAD_SOUTH_WEST:
            drawGradientDiagonal(buff, color1, color2, type == GRAD_SOUTH_WEST);
            break;
        case GRAD_CIRCLE:
        case GRAD_SQUARE:
            drawGradientCenter(buff, color1, color2, type == GRAD_SQUARE);
            break;
        case GRAD_EAST:
        case GRAD_SOUTH:
        case GRAD_NORTH_EAST:
        case GRAD_SOUTH_EAST:
        case GRAD_CIRCLE_REV:
        case GRAD_SQUARE_REV:
            drawGradient(buff, color2, color1, (Gradient) (255-type));
            break;
    }
}

void Painter::drawText(uint8_t buff[FRAME_SIZE], const Text& text) {
    if (!text.overlay) {
        // fill with background color; the text color will be overwritten later
        fill(buff, text.bgColor);
    }

    // calculate where this matrix's pixels must be colored
    uint8_t txtMask[MATRIX_PIXEL_BYTES];
    text.drawText(txtMask);

    // actually color the pixels
    for (uint8_t y = 0; y < MATRIX_ROWS; ++y) {
        for (uint8_t x = 0; x < MATRIX_COLS; ++x) {
            uint8_t d = txtMask[y * MATRIX_COLS / 8 + x / 8];
            if ((d << (x % 8)) & 0x80) {
                //pixel==1: must color it
                setPixel(text.color[R], buff, R, x, y);
                setPixel(text.color[G], buff, G, x, y);
                setPixel(text.color[B], buff, B, x, y);
            }
        }
    }
}

void Painter::drawGradientStraight(uint8_t buff[FRAME_SIZE],
                                   const uint8_t color1[3], const uint8_t color2[3], bool vert) {
    
    const uint8_t N = vert ? MATRIX_ROWS-1 : MATRIX_COLS-1; // number of shades
    const uint8_t M = vert ? MATRIX_COLS-1 : MATRIX_ROWS-1; // pixels per shade
    int16_t delta[3];
    int8_t rnd[3]; // small correction for rounding
    for (uint8_t c = 0; c < 3; c++) {
        delta[c] = (int16_t) color2[c] - color1[c];
        rnd[c] = (delta[c] >= 0) ? N / 2 : -N / 2;
    }

    for (uint8_t i = 0; i <= N; i++) {
        uint8_t col[3];

        for (uint8_t c = 0; c < 3; c++) {
            // (... + rnd) => almost like nearest rounding, instead of rounding towards zero
            col[c] = color1[c] + (delta[c] * i + rnd[c]) / N;
        }

        if (vert) {
            drawLineH(buff, 0, M, i, col);
        } else {
            drawLineV(buff, i, 0, M, col);
        }
    }
}

void Painter::drawGradientDiagonal(uint8_t buff[FRAME_SIZE],
                                   const uint8_t color1[3], const uint8_t color2[3], bool upward) {
    
    // If true, the gradient is stretched to fill the matrix (the gradient is exactly along the diagonal)
    // this is ugly with a very stretched module (i.e. when width >> height, or viceversa) because
    // it's hardly distinguishable from a straight gradient.
    // If false, the gradient is always along a 45 degrees line (but as before, the opposite vertices
    // are the full colors; i.e. this is not equal to drawing a gradient in a square containing the
    // module and then taking only the visible part: in this case one vertices would be invisible, the full
    // color would not be reached).
    const bool stretch = false;
    
    int16_t delta[3];
    int32_t rnd[3];

    const int16_t Nx = MATRIX_COLS - 1;
    const int16_t Ny = MATRIX_ROWS - 1;

    const int32_t denom = stretch ? ((int32_t)Nx * Nx + (int32_t)Ny * Ny)
                                  : (Nx + Ny);

    for (uint8_t c = 0; c < 3; c++) {
        delta[c] = (int16_t) color2[c] - color1[c];
        rnd[c] = delta[c] >= 0 ? denom / 2 : -denom / 2;
    }

    for (uint8_t y = 0; y <= Ny; y++) {
        uint8_t yy = upward ? Ny - y : y;
        for (uint8_t x = 0; x <= Nx; x++) {
            uint8_t col[3];

            int32_t dist = stretch ? (x * Nx + yy * Ny)
                                   : (x + yy);
            for (uint8_t c = 0; c < 3; c++) {
                col[c] = color1[c] + (delta[c] * dist + rnd[c]) / denom;
            }
            drawPixel(buff, x, y, col);
        }
    }
}

void Painter::drawGradientCenter(uint8_t buff[FRAME_SIZE],
                                 const uint8_t color1[3], const uint8_t color2[3], bool square) {
    uint8_t Nx = MATRIX_COLS;
    uint8_t Ny = MATRIX_ROWS;

    int16_t delta[3];
    int16_t rnd[3];
    int16_t denom = (Nx - 1) * (Ny - 1) / 2;
    if (square) {
        denom /= 2;
    }

    for (uint8_t c = 0; c < 3; c++) {
        delta[c] = color2[c] - color1[c];
        if (square) {
            rnd[c] = delta[c] >= 0 ? denom / 2 : -denom / 2;
         } else {
            rnd[c] = delta[c] >= 0 ? denom * 2 / 3 : -denom * 2 / 3;
        }
    }

    for (uint8_t x = 0; x < Nx; x++) {
        for (uint8_t y = 0; y < Ny; y++) {
            int8_t distX = x - Nx / 2;
            int8_t distY = y - Ny / 2;
            if (distX < 0) {
                distX = -(distX + 1);
            }
            if (distY < 0) {
                distY = -(distY + 1);
            }

            int16_t dist;
            if(square){
                int16_t dx = distX * (Ny - 1) / 2;
                int16_t dy = distY * (Nx - 1) / 2;
                dist = (dx >= dy ? dx : dy); // = max(dx, dy)
            } else {
                dist = distX * (Ny - 1) / 2 + distY * (Nx - 1) / 2;
            }
            
            uint8_t color[3];

            for (uint8_t c = 0; c < 3; c++) {
                color[c] = color1[c] + (delta[c] * dist + rnd[c]) / denom;
            }

            drawPixel(buff, x, y, color);
        }
    }
}

#if COLOR_DEPTH == 2
    uint8_t Painter::packDuplColor2(uint8_t color) {
        switch (color & 0xC0) {
            default:
            case 0x00:
                return 0x00;
            case 0x40:
                return 0x55;
            case 0x80:
                return 0xAA;
            case 0xC0:
                return 0xFF;
        }
    }
#endif

void Painter::swap(uint8_t& a, uint8_t& b){
    uint8_t tmp = a;
    a = b;
    b = tmp;
}

} // namespace
