/*
 Copyright (C) 2013 Stefano Zanotti @ C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Text.h"

#include <string.h>
#include "fmatdef.h"
#include "Fonts.h"

namespace FlashMat_lite {

Text::Text() :
        overlay(false),
        fontId(0),
        font(0)
{
    // default: yellow
    color[R] = 255;
    color[G] = 255;
    color[B] = 0;
    
    // default: black
    bgColor[R] = 0;
    bgColor[G] = 0;
    bgColor[B] = 0;

    // also nullify the last byte, which can never be part of the string
    memset(str, 0, TEXT_BUFFER_SIZE + 1);

    topleft[X] = 0;
    topleft[Y] = 0;

    font = Fonts::getFont(0);
}

void Text::drawText(uint8_t mat[MATRIX_PIXEL_BYTES]) const {

    // clear the matrix (from now on, only 1's will be written)
    memset(mat, 0, MATRIX_PIXEL_BYTES);

    if (!font) {
        // no font settings: nothing to do
        // NOTE: str can't be null; it's an array, not a pointer
        return;
    }

    const Font* font = this->font;

    int16_t x = topleft[X]; // the cursor
    int16_t y = topleft[Y]; // the cursor
    int16_t xst, xend; // to define where the character bitmask is
    int16_t yst, yend; // to define where the character bitmask is
    
    yst = y;
    yend = yst + 8 - 1;

    if (yend < 0 || yst >= MATRIX_ROWS) {
        return;
    }
    
    for (const uint8_t* cp = (const uint8_t*) str; ; ++cp) {
        uint8_t c = *cp;
        if (c == '\0') {
            // reached end of the string
            return;
        }

        if (c == '\n') {
            c = ' ';
        }

        xst = x - font->x0(c);
        xend = xst + 8 - 1;

        if (xst < MATRIX_COLS && xend >= 0) {
            drawChar(c, mat, x, y);
        }

        const uint8_t cw = font->width(c);
        if(cw > 0) {
            x += cw + 1;
        }
    }
}

void Text::drawChar(uint8_t c, uint8_t mat[MATRIX_PIXEL_BYTES], int8_t x,
        int8_t y) const {
            
    uint8_t mask[8];
    font->getMask(c, mask);
    
    x -= font->x0(c);
    
    // iterate over rows
    for (uint8_t i = 0; i < 8; ++i, ++y) {
        if (y >= MATRIX_ROWS) {    // we're below the window
            return;
        } else if (y < 0) {    // we're above the window
            continue;
        }

        //iterate over columns
        int8_t xx = x;
        for (uint8_t j = 0; j < 8; ++j, ++xx) {
            if (xx >= MATRIX_COLS) {    // we're to the right of the window
                break;
            } else if (xx < 0) {    // we're to the left the window
                continue;
            }

            // i * maskWidth : jump to proper mask row
            // j / 8 : jump to proper column byte
            // j % 8 : select proper bit inside the column byte (0=leftmost)
            // 0x80 >> : mask to get the proper bit
            uint8_t bit = mask[i + j / 8] & (0x80 >> (j % 8));

            if (bit) {
                // MATRIX_COLS / 8 : jump to the proper row
                // xx / 8 : jump to proper column byte
                // xx % 8 : select proper bit inside the column byte (0=leftmost)
                // 0x80 >> : mask to get the proper bit
                mat[y * (MATRIX_COLS / 8) + xx / 8] |= (0x80 >> (xx % 8));
            }
        }
    }
}

void Text::getStringPixelSize(int16_t* width, int16_t* height) const {
    if(!height && !width)
        return;
        
    if(!font) {
        if(width)
            *width = 0;
        if(height)
            *height = 0;
        return;
    }
    
    int16_t w = 0;
    const char* s = str;
    char c;
    for(;;) {
        c = *(s++);
        if (c == '\0') {
            break; // END OF LOOP
        } else {
            const uint8_t cw = font->width(c == '\n' ? ' ' : c);
            if (cw > 0) {
                w += cw + 1;
            }
        }
    }
    
    // we added one spacing too much, after the last character of each row (but only if the string wasn't empty!)
    // and after the last row: correct it (but only if the row/string wasn't empty, made only of invisible characters)
    if(str[0] == '\0'){
        *height = 0;
        *width = 0;
    } else {
        if(height)
            *height = w ? 8 : 0;
        if(width)
            *width = w ? w - 1 : w;
    }
}

void Text::center() {
    int16_t w, h;
    getStringPixelSize(&w, &h);
    topleft[X] = (MATRIX_COLS - w) / 2;
    topleft[Y] = (MATRIX_ROWS - h) / 2;
}

} //namespace
