/*
 Copyright (C) 2013 Stefano Zanotti @ C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef TEXT_H_
#define TEXT_H_

#include "fmatdef.h"
#include <inttypes.h>
#include "Font.h"

namespace FlashMat_lite {

class Text {
public:
	// The string of text to be shown
	char str[TEXT_BUFFER_SIZE + 1];
	// The foreground color
	uint8_t color[3];
	// The background color
	uint8_t bgColor[3];
	// The origin of the entire string
	int16_t topleft[2];

	// True if the text is to be drawn over the current image
	// False if the text is to be drawn over a uniform bgColor background
	bool overlay;

    uint8_t fontId; // just a nice place to hold the id for the user (the cell doesn't need it), to send it over through the commander
	const Font* font;

	Text();

	/*
	 * Given the current text/color/position/..., calculates
	 * which pixels of the RxC window must be colored (since they are part
	 * of the characters).
	 * The given matrix is row-by-row, and each bit is set to
	 * 1 (text) or 0 (background) by this function.
	 *
	 * It is then up to the caller to color its frame buffer accordingly.
	 * (it is not done by this class itself so that it needn't the implementation
	 * details of the frame buffer).
	 * It follows that color and bgColor are not used by this class, but still
	 * they are stored here to have all text-related settings collected together.
     *
     * NOTE: this function draws a "black-and-white" image: it cannot be used
     * on the frame buffer directly!
	 */
	void drawText(uint8_t mat[MATRIX_PIXEL_BYTES]) const;


	/*
	 * Helper for drawText(). Draws the given char (as defined by the font)
	 * into the given matrix (black-and-white). The char is drawn starting from
	 * topleft=[x,y], according to the x0 of the character in the current font.
	 * Checks are performed so that only those pixels of the characters
	 * that fit in the RxC matrix are drawn. This RxC is assumed to begin at
	 * topleft=[0,0].
	 * Only 1-valued pixels are written. 0-valued pixels are left unchanged,
     * so that you can draw multiple chars on the same matrix (e.g. characters
     * side-by-side in a word).
     *
     * NOTE: this function draws a "black-and-white" image: it cannot be used
     * on the frame buffer directly!
	 */
	void drawChar(uint8_t c, uint8_t mat[MATRIX_PIXEL_BYTES], int8_t x, int8_t y) const;
            
    /*
     * Calculates the space occupation of the text with all the
     * current settings (of course, 'topleft' doesn't matter).
     *
     * If you don't need an argument, provide a null argument.
     */
    void getStringPixelSize(int16_t* width, int16_t* height) const;
    
    /*
     * Changes the text position so that it is horizontally and vertically centered.
     */
    void center();
};

}
#endif /* TEXT_H_ */
