/*
 Copyright (C) 2013 Stefano Zanotti.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef BITWRITE_H_
#define BITWRITE_H_

#include <inttypes.h>

/*
 * Helps accessing single bits of a 1-byte register
 * (often used to manage settings for interrupts, timers, I/O,...)
 *
 * Usage example:
 * BSET1(TCCR1B, _BV(CS12))  // expands to SET1(TCCR1B,1 << CS12)
 *                           // sets to 1 the CS12 bit (prescaler settings)
 *                           // for the TIMER1
 * BSET0(PORTD, 0x81)	     // Clears the first and last I/O pins of PORTD
 * BGET(PORTC, _BV(PC2))     // Reads the 3rd LSB of PORTC
 * BGET(PORTB, 0xFF)         // Checks if ANY of the PORTB pins is high
 */

// set to 1 the given bit(s) of the given register
inline void BSET1(volatile uint8_t& reg, uint8_t bitmask) {
	reg |= bitmask;
}

// set to 0 the given bit(s) of the given register
inline void BSET0(volatile uint8_t& reg, uint8_t bitmask) {
	reg &= ~bitmask;
}

// set the given bit(s) of the given register
inline void BSET(bool val, volatile uint8_t& reg, uint8_t bitmask) {
	val ? BSET1(reg, bitmask) : BSET0(reg, bitmask);
}

// toggle the given bit(s) of the given register
inline void BTGL(volatile uint8_t& reg, uint8_t bitmask) {
	reg ^= bitmask;
}

// Returns true if any of the masked bits is 1
inline bool BGET(volatile uint8_t& reg, uint8_t bitmask) {
	return reg & bitmask;
}

// Versions for non-volatile byte variables

// set to 1 the given bit(s) of the given register
inline void BSET1(uint8_t& reg, uint8_t bitmask) {
    reg |= bitmask;
}

// set to 0 the given bit(s) of the given register
inline void BSET0(uint8_t& reg, uint8_t bitmask) {
    reg &= ~bitmask;
}

// set the given bit(s) of the given register
inline void BSET(bool val, uint8_t& reg, uint8_t bitmask) {
    val ? BSET1(reg, bitmask) : BSET0(reg, bitmask);
}

// toggle the given bit(s) of the given register
inline void BTGL(uint8_t& reg, uint8_t bitmask) {
    reg ^= bitmask;
}

// Returns true if any of the masked bits is 1
inline bool BGET(uint8_t& reg, uint8_t bitmask) {
    return reg & bitmask;
}

#endif /* BITWRITE_H_ */
