/*
 Copyright (C) 2013 Stefano Zanotti @ C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef FMATDEF_H_
#define FMATDEF_H_

#include <inttypes.h>

/*
 * Settings for FlashMat hardware/software
 */

namespace FlashMat_lite {

// Supported hardware (do not modify)
#define BOARD_RAINBOWDUINO_V3   1
#define BOARD_FLASHMAT          2

#define LED_MATRIX_BSRG         1  // same as Rainbowduino
#define LED_MATRIX_BRGB         2  // R and B are swapped
// NOTE: we mainly use BRGB matrices; since the blue leds (for the BRGB matrices) appear brighter
// than the red and green ones, we compensated for them in hardware (see pages 2 and 10 on the
// MY9221 datasheet, about R_EXT and current settings).
// If you use other matrices, you probably need to change these resistors too.
// Resistors are: R1 (and corresponding ones) for Blue, R2 for Green, R3 for Red (for matrices other than BRGB,
// they may correspond to different colors).
// Note that the Rainbowduino hardware does not use this compensation, i.e. it uses equal resistors for all colors.

////////////////////////////
// SETTINGS
//
// Choose your hardware here
#define USED_BOARD        BOARD_FLASHMAT
#define LED_MATRIX_TYPE   LED_MATRIX_BRGB
//
// Define the color depth for the frame buffers
// more depth -> more RAM needed
// less -> a bit more conversion needed when storing data (and maybe when driving leds)
// (this is the depth on the FlashMat; the master can use another one, and let the Commander
// do the conversions before sending the packets)
//
// RAM requirements (only for the frame buffers; then there is the space needed by other data)
//            8x8 matrix  |  8x32 matrix
// depth 8 :  384 bytes   |  1536 bytes
// depth 4 :  192 bytes   |   768 bytes
// depth 2 :   96 bytes   |   384 bytes
// depth 1 :   48 bytes   |   192 bytes
#define COLOR_DEPTH  4   // #bits per color (1, 2, 4 or 8)

////////////////////////////

// pixel-to-matrix-pin mapping
#define PINMAP_RAINBOWDUINO 1 // in short: one MY9221 for reds and half greens, the other for blues and half greens
#define PINMAP_FLASHMAT     2 // in short: MY9221 has R/G/B-dedicated outputs, use them appropriately

// Hardware-specific settings
#if USED_BOARD == BOARD_RAINBOWDUINO_V3
    #define MATRIX_ROWS              8	 // #rows in the physical RGB matrix (should be a multiple of 8)
    #define MATRIX_COLS              8   // #cols in the physical RGB matrix (should be a multiple of 8)
    #define TEXT_BUFFER_SIZE       255   // the max number of characters that can be stored as text (text is first sent to cell, then shown with specified color/position/font/...)
    #define PIN_MAPPING        PINMAP_RAINBOWDUINO
    #define USE_GCKI             false   // use the internal oscillator of MY9221, not an external grayscale clock
    
#elif USED_BOARD == BOARD_FLASHMAT
    #define MATRIX_ROWS              8   // #rows in the physical RGB matrix (should be a multiple of 8)
    #define MATRIX_COLS             32   // #cols in the physical RGB matrix (should be a multiple of 8)
    #define TEXT_BUFFER_SIZE       255   // the max number of characters that can be stored as text (text is first sent to cell, then shown with specified color/position/font/...)
    #define PIN_MAPPING        PINMAP_FLASHMAT
    #define USE_GCKI              true   // use an external grayscale clock, not the internal oscillator of MY9221
    
#else
    #error "The specified board is unknown"
#endif

// Define the RGB order in the frame buffers
// This is independent of the order in which the colors must be sent to
// the MY9221 drivers (which depends on the hardware only)
#if USED_BOARD == BOARD_RAINBOWDUINO_V3
    #define R 2
    #define G 1
    #define B 0
#elif USED_BOARD == BOARD_FLASHMAT
    #define R 0
    #define G 1
    #define B 2
#endif

// frame buffer constants
#define FRAME_SIZE_FOR(depth)          (MATRIX_ROWS * MATRIX_COLS * 3 * depth / 8)
#define FRAME_SIZE_1_COLOR_FOR(depth)  (MATRIX_ROWS * MATRIX_COLS     * depth / 8)
#define FRAME_SIZE          FRAME_SIZE_FOR(COLOR_DEPTH)          // size in bytes of the frame buffer
#define FRAME_SIZE_1_COLOR  FRAME_SIZE_1_COLOR_FOR(COLOR_DEPTH)  // as above, but for just 1 color
#define MATRIX_PIXELS       (MATRIX_ROWS * MATRIX_COLS)          // the number of pixels in the matrix
#define MATRIX_PIXEL_BYTES  (MATRIX_PIXELS / 8)                  // the number of bytes needed to store 1bit per pixel (pixels are "packed" 8-by-8) (in 1 color plane; this is used for masks, not for RGB frames at color depth = 1)
#define MATRIX_ROW_SIZE     (MATRIX_COLS * COLOR_DEPTH   / 8)    // the number of bytes necessary to store 1 row of 1 color

// Constants to access properties on X and Y axis (font spacing, etc.)
#define X 0
#define Y 1

// Utilities

#define MAKE_RGB(r, g, b)   { (R == 0) * (uint8_t)(r) + (G == 0) * (uint8_t)(g) + (B == 0) * (uint8_t)(b), \
                              (R == 1) * (uint8_t)(r) + (G == 1) * (uint8_t)(g) + (B == 1) * (uint8_t)(b), \
                              (R == 2) * (uint8_t)(r) + (G == 2) * (uint8_t)(g) + (B == 2) * (uint8_t)(b) }

#define MAKE_XY_u8(x, y)    { (X == 0) * (x) + (Y == 0) * (y), \
                              (X == 1) * (x) + (Y == 1) * (y) }
                                                  
#define MAKE_XY_i16(x, y)   { (X == 0) * (x) + (Y == 0) * (y), \
                              (X == 1) * (x) + (Y == 1) * (y) }

//////////////////////////////////////////////////////////////
// Consistency checks
//////////////////////////////////////////////////////////////

#if COLOR_DEPTH != 1 && COLOR_DEPTH != 2 && COLOR_DEPTH !=4 && COLOR_DEPTH !=8
    #error "COLOR_DEPTH must be one of 1,2,4,8"
#endif

#if (R != 0 && G != 0 && B != 0) || (R != 1 && G != 1 && B != 1) || (R != 2 && G != 2 && B != 2) || (R + G + B != 3)
    #error "R,G,B must be the numbers 0,1,2 (in any order)"
#endif

#if (X != 0 && Y != 0) || (X != 1 && Y != 1) || (X + Y != 1)
    #error "X,Y must be the numbers 0,1 (in any order)"
#endif

#if MATRIX_COLS <= 0 || MATRIX_COLS > 255 || (MATRIX_COLS & 0x07) != 0
    #error "MATRIX_COLS must be a multiple of 8, between 8 and 248"
#endif

#if MATRIX_ROWS <= 0 || MATRIX_ROWS > 255 || (MATRIX_ROWS & 0x07) != 0
    #error "MATRIX_ROWS must be a multiple of 8, between 8 and 248"
#endif

#if PIN_MAPPING != PINMAP_RAINBOWDUINO && PIN_MAPPING != PINMAP_FLASHMAT
    #error "Unknown PIN_MAPPING"
#endif

#if USED_BOARD != BOARD_RAINBOWDUINO_V3 && USED_BOARD != BOARD_FLASHMAT
    #error "Unknown USED_BOARD"
#endif

#if LED_MATRIX_TYPE != LED_MATRIX_BSRG && LED_MATRIX_TYPE != LED_MATRIX_BRGB
    #error "Unknown LED_MATRIX_TYPE"
#endif

#if TEXT_BUFFER_SIZE > 255 || TEXT_BUFFER_SIZE < 0
    #error "TEXT_BUFFER_SIZE must fit into a uint8_t"
#endif

} // namespace

#endif /* FMATDEF_H_ */
