/*
 Copyright (C) 2013 Stefano Zanotti @ C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef TIME_UTILS_H_
#define TIME_UTILS_H_

namespace FlashMat_lite {
    // NOTE depending on the clock frequency, you can only get some amounts of delay:
    // + 64MHz  -> 0us ~ 4095us   with "precision" 1/16us
    // + 32MHz  -> 0us ~ 8191us   with "precision" 1/8us
    // + 16MHz  -> 0us ~ 16383us  with "precision" 1/4us
    // +  8MHz  -> 0us ~ 32767us  with "precision" 1/2us
    // +  4MHz  -> 0us ~ 32767us  with resolution  1us
    // +  2MHz  -> 0us ~ 32767us  with resolution  2us
    // +  1MHz  -> 0us ~ 32767us  with resolution  4us
    // Actual delays may be rounded to floor.
    //
    // NOTE: the actual timings rely on this function call being optimized, so that
    // the manipulation of 'us' is done at compile time.
    // This might not be true if you provide a variable as argument; it's better
    // to provide a constant (after all, this function is made for small and tweaked
    // delays, so they should be well known at compile time).
    
    inline void delayUs(uint16_t us) __attribute__((always_inline));
    
    inline void delayUs(uint16_t us) {

        // all these shifts are optimized away
    #if F_CPU == 64000000UL
        us <<= 4;
    #elif F_CPU == 32000000UL
        us <<= 3;
    #elif F_CPU == 16000000UL
        us <<= 2;
    #elif F_CPU == 8000000UL
        us <<= 1;
    #elif F_CPU == 4000000UL
        // do nothing
    #elif F_CPU == 2000000UL
        us >>= 1;
    #elif F_CPU == 1000000UL
        us >>= 2;
    #else
        #error "Clock frequency not yet supported"
    #endif
    
        if (us == 0) { // (optimized away)
            return;
        }
    
        // now 'us' is actually the number of 4-cycle iterations
        // remove one cycle and add two nops:
        // the nops and the two LDI resulting from the constant argument
        // compensate the missing loop.
        // Since the brne below takes 1 cycle when it does not branch (exit condition),
        // we add a third nop here.
    
        us--; // remove one cycle (optimized away)
        __asm__ __volatile__ (
            "nop" "\n\t"
            "nop" "\n\t"
    	    "nop"
    	    );
    
        // busy wait
        __asm__ __volatile__ (
    	    "1: sbiw %0,1" "\n\t" // 2 cycles
    	    "brne 1b" : "=w" (us) : "0" (us) // 2 cycles
    	    );
    }
    
    inline void nop() __attribute__((always_inline));
    
    inline void nop() {
        __asm__ __volatile__ ("nop");
    }
}

#endif /* TIME_UTILS_H_ */